package com.gmei.up.utils

import scala.collection.mutable.ListBuffer
import scala.util.Try

import org.json4s.JsonDSL._
import org.json4s.jackson.JsonMethods._
import org.json4s.JObject

import org.apache.http.HttpHost
import org.apache.http.auth.{ AuthScope, UsernamePasswordCredentials }
import org.apache.http.impl.client.BasicCredentialsProvider
import org.apache.http.impl.nio.client.HttpAsyncClientBuilder

import org.elasticsearch.client.RestClient
import org.elasticsearch.client.RestClientBuilder.HttpClientConfigCallback
import com.sksamuel.elastic4s.requests.searches.{ SearchRequest, SearchResponse }
import com.sksamuel.elastic4s.requests.searches.sort.{ FieldSort, SortOrder, ScriptSortType }
import com.sksamuel.elastic4s.ElasticDsl._
import com.sksamuel.elastic4s.{ HitReader, Hit, ElasticClient }
import com.sksamuel.elastic4s.http.JavaClient
import com.sksamuel.elastic4s.requests.searches.queries.Query
import com.sksamuel.elastic4s.requests.searches.queries.term.TermsQuery
import com.sksamuel.elastic4s.requests.script.{ Script, ScriptType }
import com.sksamuel.elastic4s.requests.searches.queries.funcscorer.{
  CombineFunction,
  FunctionScoreQueryScoreMode,
  WeightScore,
  ScoreFunction
}

object ESClient {
  private class ElasticCredentials(username: String, password: String) extends HttpClientConfigCallback {
    val credentials = new UsernamePasswordCredentials(username, password)
    val credentialsProvider = new BasicCredentialsProvider()
    credentialsProvider.setCredentials(AuthScope.ANY, credentials)

    override def customizeHttpClient(httpClientBuilder: HttpAsyncClientBuilder): HttpAsyncClientBuilder =
      httpClientBuilder.setDefaultCredentialsProvider(credentialsProvider)
  }

  def create(hostname: String, port: Int, username: String, password: String): ElasticClient = {
    val restClient = RestClient
      .builder(new HttpHost(hostname, port, "http"))
      .setHttpClientConfigCallback(new ElasticCredentials(username, password))
      .build()

    ElasticClient(JavaClient.fromRestClient(restClient))
  }
}

object ES {
  case class Character(id: String, index: String, projects: String)

  implicit object CharacterHitReader extends HitReader[Character] {
    override def read(hit: Hit): Try[Character] = {
      val source = hit.sourceAsMap
      Try(Character(source("id").toString, hit.index, source("tags_v3").toString))
    }
  }

  def test(diaryQuery: SearchRequest, tractateQuery: String) = {
    // TODO read from config
    val client: ElasticClient = ESClient.create("172.16.52.33", 9200, "elastic", "gengmei!@#")

    val bb = multi(
      diaryQuery,
      search("gm-dbmw-tractate-read").source(tractateQuery)
    )

    println(bb.show)
    println(diaryQuery)
    println(tractateQuery)

    // TODO remove await
    val resp = client.execute(bb).await

    // println(resp)

    val a = resp.result.to[Character]

    println("%%%%%%%%%%%")
    println(a.size)
    a.foreach { x =>
      println(x.id, x.index, x.projects)
    }
    println("%%%%%%%%%%%")

    resp
  }

  def generateDiaryQuery(
      projects: List[String],
      secondDemands: List[String],
      cityId: Int = -1
  ): SearchRequest = {
    val size = 15
    val includes = List("id", "tags_v3")
    var scoreListBuffer: ListBuffer[ScoreFunction] = ListBuffer.empty[ScoreFunction]
    var shouldListBuffer: ListBuffer[Query] = ListBuffer.empty[Query]

    if (projects.size > 0) {
      scoreListBuffer += WeightScore(60).filter(
        must(termsQuery("tags_v3", projects), termQuery("is_promote", true))
      )
      scoreListBuffer += WeightScore(50).filter(termsQuery("tags_v3", projects))
      shouldListBuffer += TermsQuery("tags_v3", projects)
    }
    if (secondDemands.size > 0) {
      scoreListBuffer += WeightScore(40).filter(termsQuery("second_demands", secondDemands))
      shouldListBuffer += TermsQuery("second_demands", secondDemands)
    }

    val res = search("gm-dbmw-diary-read")
      .query {
        functionScoreQuery()
          .functions(scoreListBuffer.toList)
          .scoreMode(FunctionScoreQueryScoreMode.Max)
          .boostMode(CombineFunction.Replace)
          .query {
            boolQuery()
              .must(
                termQuery("is_online", true),
                termQuery("has_cover", true),
                termQuery("is_sink", false),
                termQuery("has_after_cover", true),
                termQuery("has_before_cover", true),
                rangeQuery("content_level").gte(3),
                termQuery("content_simi_bol_show", 0)
              )
              .not(termQuery("show_by_index", 2))
              .should(shouldListBuffer.toList)
              .minimumShouldMatch(1)
          }
      }
      .size(size)
      .sourceInclude(includes)
      .sortBy(
        scriptSort(
          script(
            "if (doc['doctor.hospital.city_tag_id'].size()!=0){if (doc['doctor.hospital.city_tag_id'].value ==params.user_city_tag_id){return 2;} else if (doc['nearby_city_tags_v1.tag_id'].contains(params.user_city_tag_id) ) {return 1;} else {return 0;} }else{return 0}"
          ).params(Map("user_city_tag_id" -> cityId)).lang("painless")
        ).typed("number").order(SortOrder.Desc),
        FieldSort("_score").desc(),
        FieldSort("has_video_cover").asc(),
        FieldSort("offline_score").desc(),
        FieldSort("good_click").desc(),
        FieldSort("last_update_time").desc()
      )

    println("diary query")
    println(res.show)

    res
  }

  def generateTractateQuery(projects: List[String], secondDemands: List[String]): String = {
    val size = 30
    val includes = List("id", "tags_v3")

    val functions = new ListBuffer[JObject]()
    val shoulds = new ListBuffer[JObject]()
    if (projects.size > 0) {
      val promotes: List[JObject] =
        List(("terms" -> ("tags_v3" -> projects)), ("term" -> ("is_promote" -> true)))

      functions += ("filter" -> ("bool" -> ("must" -> promotes))) ~ ("weight" -> 60)

      functions += ("filter" -> ("terms" -> { "tags_v3" -> projects })) ~ ("weight" -> 50)

      shoulds += ("terms" -> ("tags_v3" -> projects))
    }

    if (secondDemands.size > 0) {
      functions += ("filter" -> ("terms" -> ("second_demands" -> secondDemands))) ~ ("weight" -> 40)

      shoulds += ("terms" -> ("second_demands" -> secondDemands))
    }

    val query = s"""
{
  "query": {
    "function_score": {
      "functions": ${pretty(render(functions.toList))},
      "boost_mode": "replace",
      "score_mode": "max",
      "query": {
        "bool": {
          "should": ${pretty(render(shoulds.toList))},
          "minimum_should_match": 1,
          "must": [
            {
              "term": {
                "is_online": "true"
              }
            },
            {
              "terms": {
                "content_level": [6, 5, 4, 3.5, 3]
              }
            }
          ],
          "must_not": [
            { "term": { "status": 4 } },
            { "term": { "show_by_index": 2 } }
          ]
        }
      }
    }
  },
  "size": ${size},
  "_source": {"include": ${pretty(render(includes))}}
}
""".stripMargin

    // println("tractate query:")
    // println(query)
    query
  }
}
